"""Optimizers for generated python code."""

import ast
from ast import AST
from ast import Assign as AstAssign
from ast import Load as AstLoad
from ast import Module as AstModule
from ast import Name as AstName
from ast import Store as AstStore
from ast import expr as AstExpr
from ast import parse as ast_parse
from collections import defaultdict, deque
from itertools import chain
from typing import Any, Callable, Dict, List, Optional, Tuple

from ._utils import PY_VERSION, ast_unparse


def ast_are_fuzzy_equal(left, right, fuzzy_cmp, fields_to_skip=frozenset()):
    cls_left = type(left)
    if issubclass(cls_left, AST):
        if cls_left is not type(right):
            return False

        if fuzzy_cmp(left, right):
            return True

        for right_field in right._fields:  # pylint: disable=protected-access
            if right_field not in fields_to_skip:
                right_value_or_values = getattr(right, right_field)
                left_value_or_values = getattr(left, right_field)
                if isinstance(right_value_or_values, list):
                    length = len(right_value_or_values)
                    if length != len(left_value_or_values):
                        return False
                    index = 0
                    while index < length:
                        if not ast_are_fuzzy_equal(
                            left_value_or_values[index],
                            right_value_or_values[index],
                            fuzzy_cmp,
                            fields_to_skip,
                        ):
                            return False
                        index += 1

                elif not ast_are_fuzzy_equal(
                    left_value_or_values,
                    right_value_or_values,
                    fuzzy_cmp,
                    fields_to_skip,
                ):
                    return False

        return True

    else:
        return left == right

    # if type(left) is not type(right):
    #     return False
    # _stack = [(left, right)]
    # while _stack:
    #     left, right = _stack.pop()
    #     cls_left = type(left)
    #     if not issubclass(cls_left, AST):
    #         if left != right:
    #             return False
    #         continue

    #     if cls_left is not type(right):
    #         return False

    #     if fuzzy_cmp(left, right):
    #         continue

    #     for right_field in right._fields:
    #         if right_field not in fields_to_skip:
    #             left_value_or_values = getattr(left, right_field)
    #             right_value_or_values = getattr(right, right_field)
    #             if isinstance(right_value_or_values, list):
    #                 if len(left_value_or_values) != len(right_value_or_values):
    #                     return False
    #                 _stack.extend(
    #                     zip(left_value_or_values, right_value_or_values)
    #                 )
    #             else:
    #                 _stack.append(
    #                     (left_value_or_values, right_value_or_values)
    #                 )
    # return True


class AstMergeCtx:
    """Holds last used indexes for multi-statement AST nodes."""

    __slots__ = ["_body_id_to_last_used_index"]

    def __init__(self):
        self._body_id_to_last_used_index = {}

    def use_index(self, body, index):
        body_id = id(body)
        if body_id in self._body_id_to_last_used_index:
            self._body_id_to_last_used_index[body_id] = max(
                self._body_id_to_last_used_index[body_id], index
            )
        else:
            self._body_id_to_last_used_index[body_id] = index

    def get_last_used(self, body):
        body_id = id(body)
        if body_id in self._body_id_to_last_used_index:
            return self._body_id_to_last_used_index[body_id]
        return 0

    def get_index_to_insert_to(self, body):
        body_id = id(body)
        if body_id in self._body_id_to_last_used_index:
            self._body_id_to_last_used_index[body_id] += 1
        else:
            self._body_id_to_last_used_index[body_id] = 0
        return self._body_id_to_last_used_index[body_id]


def ast_merge(
    left_body,
    right,
    fuzzy_cmp,
    ctx=None,
    _attrs_with_stmts=frozenset(["body", "orelse", "handlers", "finalbody"]),
):
    if ctx is None:
        ctx = AstMergeCtx()

    if isinstance(left_body, AstModule):
        left_body = left_body.body

    if isinstance(right, AST):
        if not hasattr(right, "body"):
            left_body.insert(ctx.get_index_to_insert_to(left_body), right)
            return

        if isinstance(right, AstModule):
            for right_stmt in right.body:
                ast_merge(left_body, right_stmt, fuzzy_cmp, ctx)
            return

        left = None
        index_ = ctx.get_last_used(left_body)
        length = len(left_body)
        while index_ < length:
            left_stmt = left_body[index_]
            if ast_are_fuzzy_equal(
                left_stmt,
                right,
                fuzzy_cmp,
                fields_to_skip=_attrs_with_stmts,
            ):
                left = left_stmt
                ctx.use_index(left_body, index_)
                break
            index_ += 1

        if left is None:
            left_body.insert(ctx.get_index_to_insert_to(left_body), right)
        else:
            for (
                right_field
            ) in right._fields:  # pylint: disable=protected-access
                if right_field in _attrs_with_stmts:
                    for value in getattr(right, right_field):
                        ast_merge(
                            getattr(left_stmt, right_field),
                            value,
                            fuzzy_cmp,
                            ctx,
                        )

    else:
        raise AssertionError


def node_from_node_path(path):
    if isinstance(path[1], int):
        return path[0][path[1]]
    return getattr(path[0], path[1])


def replace_node_by_node_path(path, new_node):
    if isinstance(path[1], int):
        path[0][path[1]] = new_node
    else:
        setattr(path[0], path[1], new_node)


class ExprInfo:
    """ast expression usage info."""

    __slots__ = [
        "node_paths",
        "parents",
        "children",
        "number",
    ]

    def __init__(self):
        self.node_paths = []
        self.parents = set()
        self.children = set()
        self.number = 0

    def to_dict(self):  # pragma: no cover
        return {
            "node_paths": self.node_paths,
            "parents": self.parents,
            "children": self.children,
            "number": self.number,
        }


class CodeLayer:
    """Stores info about ast node bodies (lists of statements)."""

    __slots__ = [
        "body",
        "stack_indexes",
        "stack_expr_code_to_info",
        "children",
        "number_updates",
    ]

    def __init__(self, body):
        self.body = body
        self.stack_indexes = [0]
        self.stack_expr_code_to_info: "List[Optional[Dict[str, ExprInfo]]]" = [
            None
        ]
        self.children = []
        self.number_updates = []

    def to_dict(self):  # pragma: no cover
        return {
            "body": self.body,
            "stack_indexes": self.stack_indexes,
            "stack_expr_code_to_info": [
                {
                    expr_code: info.to_dict()
                    for expr_code, info in (item and item.items() or ())
                }
                for item in self.stack_expr_code_to_info
            ],
            "children": [child.to_dict() for child in self.children],
            "number_updates": self.number_updates,
        }

    def start_tracking_number_updates(self):
        self.number_updates.append(defaultdict(int))

    def stop_tracking_number_updates_and_pop_deltas(self):
        return self.number_updates.pop()

    def track_expr_chain(self, expr_code_tree, stack_index):
        number_updates = (
            self.number_updates[-1] if self.number_updates else None
        )

        for index in range(len(self.stack_indexes) - 1, -1, -1):
            if self.stack_indexes[index] > stack_index:
                continue

            expr_code_to_info = self.stack_expr_code_to_info[index]
            if expr_code_to_info is None:
                expr_code_to_info = self.stack_expr_code_to_info[index] = (
                    defaultdict(ExprInfo)
                )
            for expr_code, l_expr_info in expr_code_tree.items():
                if expr_code in expr_code_to_info:
                    expr_info = expr_code_to_info[expr_code]
                    expr_info.node_paths.extend(l_expr_info.node_paths)
                    expr_info.parents.update(l_expr_info.parents)
                    expr_info.children.update(l_expr_info.children)
                    expr_info.number += l_expr_info.number
                else:
                    expr_code_to_info[expr_code] = l_expr_info
                if number_updates is not None:
                    number_updates[(index, expr_code)] += l_expr_info.number

            return
        raise AssertionError

    def split_current_layer(self, index):
        if not self.stack_indexes or self.stack_indexes[-1] != index:
            self.stack_indexes.append(index)
            self.stack_expr_code_to_info.append(None)


class NewLayerCtx:
    """Manages code layers of OptimizationStage1."""

    __slots__ = ["visitor", "code_layer", "prev"]

    def __init__(self, visitor, code_layer):
        self.visitor = visitor
        self.code_layer = code_layer

    def __enter__(self):
        visitor = self.visitor
        self.prev = visitor.parent_layer, visitor.current_layer
        visitor.current_layer.children.append(self.code_layer)
        visitor.parent_layer = visitor.current_layer
        visitor.current_layer = self.code_layer
        visitor.stack_body_indexes.append(0)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.visitor.stack_body_indexes.pop()
        self.visitor.parent_layer, self.visitor.current_layer = self.prev


class NoTrackNumbersCtx:
    """Stops OptimizationStage1 expr number counting."""

    __slots__ = [
        "visitor",
        "prev_mode_track_numbers",
    ]

    def __init__(self, visitor):
        self.visitor = visitor
        self.prev_mode_track_numbers = None

    def __enter__(self):
        visitor = self.visitor
        self.prev_mode_track_numbers = visitor.mode_track_numbers
        visitor.mode_track_numbers = False

    def __exit__(self, exc_type, exc_value, exc_traceback):
        visitor = self.visitor
        visitor.mode_track_numbers = self.prev_mode_track_numbers


class ExprChainCollector:
    """Snapshots OptimizationStage1 found expression chains."""

    __slots__ = [
        "visitor",
        "prev_mode_collect_expr_chain",
        "prev_parent_expr_code",
        "prev_expr_code_tree",
        "expr_code_tree",
    ]

    def __init__(self, visitor):
        self.visitor = visitor

    def start(self):
        visitor = self.visitor
        if visitor.mode_collect_expr_chain:
            raise AssertionError("bug")

        self.prev_mode_collect_expr_chain = visitor.mode_collect_expr_chain

        self.prev_parent_expr_code = visitor.parent_expr_code
        self.prev_expr_code_tree = visitor.expr_code_tree

        visitor.mode_collect_expr_chain = True
        visitor.parent_expr_code = None
        visitor.expr_code_tree = defaultdict(ExprInfo)

    def stop(self):
        visitor = self.visitor
        visitor.mode_collect_expr_chain = self.prev_mode_collect_expr_chain

        self.expr_code_tree = visitor.expr_code_tree

        visitor.parent_expr_code = self.prev_parent_expr_code
        visitor.expr_code_tree = self.prev_expr_code_tree


class OptimizationStage1(ast.NodeVisitor):
    """Common sub-expression elimination by ast node visitor."""

    __slots__ = [
        "tree",
        "_consumed_exprs",
        "current_layer",
        "expr_code_tree",
        "exprs_to_optimize",
        "mode_collect_expr_chain",
        "mode_track_numbers",
        "no_side_effects_test",
        "node_path",
        "parent_expr_code",
        "parent_layer",
        "root_layer",
        "stack_body_indexes",
        "stack_expr_visibilities",
        "methods",
        "_node_code_cache",
    ]

    def __init__(self):
        self.exprs_to_optimize = set()
        self._consumed_exprs = set()
        self.methods: Dict[str, Callable] = defaultdict(
            lambda: self.generic_visit
        )
        self._node_code_cache = {}
        for name in dir(self):
            if name.startswith("visit_"):
                self.methods[name] = getattr(self, name)

    def _init_optimization_params(self, tree, no_side_effects_test):
        self.tree = tree
        self.no_side_effects_test = no_side_effects_test

        self.root_layer = CodeLayer(self.tree.body)
        self.parent_layer = self.root_layer
        self.current_layer = self.root_layer

        self.node_path: "Tuple[Any, Any]" = (None, None)
        self.stack_body_indexes = []

        self.mode_collect_expr_chain = None
        self.mode_track_numbers = True

        self.parent_expr_code = None
        self.expr_code_tree = defaultdict(ExprInfo)
        self.stack_expr_visibilities = None

    def use_expression(self, expression_code):
        if expression_code not in self._consumed_exprs:
            self._consumed_exprs.add(expression_code)
            # a and (a + 1) -> a and a + 1
            # or simply a/2 -> a / 2
            self.exprs_to_optimize.add(ast_unparse(ast_parse(expression_code)))
        return expression_code

    def run(self, tree, no_side_effects_test):
        self._init_optimization_params(tree, no_side_effects_test)
        self.raw_visit(self.tree)

        if self.root_layer is None:
            raise AssertionError("bug")

        paths_to_process: "deque[Tuple[CodeLayer, ...]]" = deque(
            [(self.root_layer,)]
        )
        while paths_to_process:
            l_path = paths_to_process.popleft()
            l_children = l_path[-1].children
            if l_children:
                for l_code_layer in l_children:
                    paths_to_process.append(l_path + (l_code_layer,))
                continue
            for index in range(len(l_path) - 1, 0, -1):
                child_expr_code_to_info = l_path[
                    index
                ].stack_expr_code_to_info[0]
                if child_expr_code_to_info is None:
                    continue
                for expr_code, info in child_expr_code_to_info.items():
                    if not info.node_paths:
                        continue
                    for parent_index in range(index):
                        parent_expr_code_to_info = l_path[
                            parent_index
                        ].stack_expr_code_to_info[0]
                        if (
                            parent_expr_code_to_info
                            and expr_code in parent_expr_code_to_info
                        ):
                            parent_info = parent_expr_code_to_info[expr_code]
                            parent_info.number += info.number
                            parent_info.node_paths.extend(info.node_paths)
                            parent_info.children.update(info.children)
                            parent_info.parents.update(info.parents)
                            info.node_paths.clear()
                            info.number = 0
                            break

        # from pprint import pprint

        # pprint(self.root_layer.to_dict())
        # breakpoint()

        code_layers_to_process = deque([self.root_layer])
        key_to_index: "Dict[Tuple[int, int], int]" = {}
        number_of_replacements = 0
        while code_layers_to_process:
            code_layer = code_layers_to_process.pop()
            for l_code_layer in code_layer.children:
                code_layers_to_process.append(l_code_layer)

            for expr_code_to_info, body_index in zip(
                reversed(code_layer.stack_expr_code_to_info),
                reversed(code_layer.stack_indexes),
            ):
                if expr_code_to_info is None:
                    continue

                for expr_code, info in sorted(
                    filter(
                        lambda item: item[1].node_paths and item[1].number > 1,
                        expr_code_to_info.items(),
                    ),
                    key=lambda item: self.get_expr_depth_below(
                        item[0],
                        expr_code_to_info,  # pylint: disable=cell-var-from-loop
                    ),
                ):
                    if info.parents and all(
                        info.number == expr_code_to_info[parent_code].number
                        for parent_code in info.parents
                    ):
                        continue

                    local_var_name = f"_r{number_of_replacements}_"
                    number_of_replacements += 1

                    replacement_node = AstAssign(
                        targets=[AstName(id=local_var_name, ctx=AstStore())],
                        value=node_from_node_path(
                            expr_code_to_info[expr_code].node_paths[0]
                        ),
                    )
                    # ast visit_Assign get_type_comment fails without it
                    replacement_node.lineno = None  # type: ignore

                    new_node = AstName(id=local_var_name, ctx=AstLoad())
                    for node_path in info.node_paths:
                        replace_node_by_node_path(node_path, new_node)

                    key = id(code_layer.body), body_index
                    if key in key_to_index:
                        key_to_index[key] = index = key_to_index[key] + 1
                    else:
                        index = key_to_index[key] = body_index

                    code_layer.body.insert(index, replacement_node)

    def get_expr_depth_below(self, expr_code, expr_code_to_info):
        stack = [(expr_code, 0)]
        max_depth = -1
        visited = {expr_code}
        while stack:
            expr, depth = stack.pop()
            children = expr_code_to_info[expr].children
            if children:
                for l_expr in children:
                    if l_expr not in visited:
                        stack.append((l_expr, depth + 1))
                        visited.add(l_expr)
            elif max_depth < depth:
                max_depth = depth
        return max_depth

    def visit(self, node):
        raise NotImplementedError

    MODE_FIRST_ONLY = -2
    MODE_ALL_BUT_FIRST = -3

    def visit_by_attr(
        self,
        node,
        attr,
        mode=None,
        _stmt_attrs=frozenset(["body", "orelse", "finalbody"]),
    ):
        value = getattr(node, attr)
        parent_expr_code = self.parent_expr_code
        if isinstance(value, list):
            if mode is None:
                index = 0
                length = len(value)
            elif mode == self.MODE_FIRST_ONLY:
                index = 0
                length = min(1, len(value))
            else:
                index = 1
                length = len(value)

            if attr in _stmt_attrs:
                while index < length:
                    self.node_path = (value, index)
                    self.stack_body_indexes.append(index)
                    item = value[index]
                    # self.raw_visit(item)
                    self.methods[f"visit_{item.__class__.__name__}"](item)
                    self.stack_body_indexes.pop()
                    index += 1

            else:
                while index < length:
                    item = value[index]
                    if isinstance(item, AST):  # pragma: no cover
                        self.node_path = (value, index)
                        # self.raw_visit(item)
                        self.methods[f"visit_{item.__class__.__name__}"](item)
                        self.parent_expr_code = parent_expr_code
                    index += 1
        elif isinstance(value, AST):
            self.node_path = (node, attr)
            # self.raw_visit(value)
            self.methods[f"visit_{value.__class__.__name__}"](value)
            self.parent_expr_code = parent_expr_code

    def raw_visit(self, node):
        self.methods[f"visit_{node.__class__.__name__}"](node)

    def generic_visit(
        self,
        node,
        _non_idempotent_expr_nodes=frozenset(
            [
                "Dict",
                "List",
                "Set",
                "DictComp",
                "ListComp",
                "SetComp",
                "GeneratorExp",
                "Call",
                "NamedExpr",
            ]
        ),
    ):
        expr_chain_collector = None
        node_name = node.__class__.__name__

        if isinstance(node, AstExpr):
            if self.mode_collect_expr_chain:
                if node_name not in _non_idempotent_expr_nodes:
                    expr_code = self.get_node_code(node)
                    expr_info = self.expr_code_tree[expr_code]
                    if self.parent_expr_code is not None:
                        self.expr_code_tree[
                            self.parent_expr_code
                        ].children.add(expr_code)
                        expr_info.parents.add(self.parent_expr_code)
                    expr_info.number += self.mode_track_numbers
                    expr_info.node_paths.append(self.node_path)

                    self.parent_expr_code = expr_code

            else:
                expr_code = self.get_node_code(node)
                if expr_code in self.exprs_to_optimize:
                    expr_chain_collector = ExprChainCollector(self)
                    expr_chain_collector.start()
                    if node_name not in _non_idempotent_expr_nodes:
                        expr_info = self.expr_code_tree[expr_code]
                        expr_info.number += self.mode_track_numbers
                        expr_info.node_paths.append(self.node_path)
                        self.parent_expr_code = expr_code
                    else:
                        self.parent_expr_code = None

        if node_name in self.CUSTOM_EXPR_VISITORS:
            self.CUSTOM_EXPR_VISITORS[node_name](self, node)

        else:
            for field in node._fields:  # pylint: disable=protected-access
                self.visit_by_attr(node, field)

        if expr_chain_collector is not None:
            expr_chain_collector.stop()
            self.current_layer.track_expr_chain(
                expr_chain_collector.expr_code_tree,
                self.stack_body_indexes[-1],
            )

    def get_node_code(self, node):
        if self.parent_expr_code is not None:
            key = self.parent_expr_code, self.node_path[-1]
            if key in self._node_code_cache:
                return self._node_code_cache[key]
            expr_code = ast_unparse(node)
            self._node_code_cache[key] = expr_code
            return expr_code
        return ast_unparse(node)

    def _custom_expr_visit_call(self, node):
        self.visit_by_attr(node, "func")
        self.visit_by_attr(node, "args")
        self.visit_by_attr(node, "keywords")
        if not self.mode_collect_expr_chain and self.may_have_side_effects(
            node
        ):
            self.split_current_layer()

    def _custom_expr_visit_list(self, node):
        self.visit_by_attr(node, "elts")

    def _custom_expr_visit_dict(self, node):
        self.visit_by_attr(node, "keys")
        self.visit_by_attr(node, "values")

    def _custom_expr_visit_comp(self, node):
        self.visit_by_attr(node.generators[0], "iter")

    def _custom_expr_visit_ifexp(self, node):
        self.visit_by_attr(node, "test")
        with NoTrackNumbersCtx(self):
            self.visit_by_attr(node, "body")
            self.visit_by_attr(node, "orelse")

    def _custom_expr_visit_named_expr(self, node):
        if self.may_have_side_effects(node.target):
            self.split_current_layer(after_current_line=True)
        self.visit_by_attr(node, "value")

    def _custom_expr_visit_bool_op(self, node):
        self.visit_by_attr(node, "values", self.MODE_FIRST_ONLY)
        with NoTrackNumbersCtx(self):
            self.visit_by_attr(node, "values", self.MODE_ALL_BUT_FIRST)

    CUSTOM_EXPR_VISITORS = {
        "Call": _custom_expr_visit_call,
        "List": _custom_expr_visit_list,
        "Set": _custom_expr_visit_list,
        "Tuple": _custom_expr_visit_list,
        "Dict": _custom_expr_visit_dict,
        "ListComp": _custom_expr_visit_comp,
        "SetComp": _custom_expr_visit_comp,
        "DictComp": _custom_expr_visit_comp,
        "GeneratorExp": _custom_expr_visit_comp,
        "IfExp": _custom_expr_visit_ifexp,
        "NamedExpr": _custom_expr_visit_named_expr,
        "BoolOp": _custom_expr_visit_bool_op,
    }

    def new_layer_ctx(self, body):
        return NewLayerCtx(self, CodeLayer(body))

    def split_current_layer(self, after_current_line=False):
        self.current_layer.split_current_layer(
            self.stack_body_indexes[-1] + 1
            if after_current_line
            else self.stack_body_indexes[-1]
        )

    def visit_If(self, node):
        if self.may_have_side_effects(node.test):
            self.visit_by_attr(node, "test")
            with self.new_layer_ctx(node.body):
                self.visit_by_attr(node, "body")

            if node.orelse:
                with self.new_layer_ctx(node.orelse):
                    self.visit_by_attr(node, "orelse")

        else:
            self.current_layer.start_tracking_number_updates()
            self.visit_by_attr(node, "body")
            body_deltas = (
                self.current_layer.stop_tracking_number_updates_and_pop_deltas()
            )
            self.current_layer.start_tracking_number_updates()
            self.visit_by_attr(node, "orelse")
            orelse_deltas = (
                self.current_layer.stop_tracking_number_updates_and_pop_deltas()
            )

            for key in set(chain(body_deltas, orelse_deltas)):
                index, expr_code = key
                d1 = body_deltas.get(key, 0)
                d2 = orelse_deltas.get(key, 0)
                expr_info = self.current_layer.stack_expr_code_to_info[index][
                    expr_code
                ]
                if d1 == 0 or d2 == 0:
                    expr_info.number -= d1 + d2
                else:
                    expr_info.number -= (d1 + d2) * 0.5

    def visit_Try(self, node):
        with self.new_layer_ctx(node.body):
            self.visit_by_attr(node, "body")
        for handler in node.handlers:
            with self.new_layer_ctx(handler.body):
                self.visit_by_attr(handler, "body")
        with self.new_layer_ctx(node.orelse):
            self.visit_by_attr(node, "orelse")
        with self.new_layer_ctx(node.finalbody):
            self.visit_by_attr(node, "finalbody")

    def visit_Constant(self, node):
        pass

    if PY_VERSION < (3, 8):

        def visit_Str(self, node):
            pass

        def visit_Num(self, node):
            pass

    def visit_Name(self, node):
        pass

    def visit_Delete(self, node):  # pylint: disable=unused-argument
        self.split_current_layer(after_current_line=True)

    def visit_While(self, node):
        with self.new_layer_ctx(node.body):
            self.visit_by_attr(node, "body")

    def visit_For(self, node):
        self.visit_by_attr(node, "iter")
        with self.new_layer_ctx(node.body):
            self.split_current_layer()
            self.visit_by_attr(node, "body")

    def visit_With(self, node):
        self.visit_by_attr(node, "items")
        with self.new_layer_ctx(node.body):
            self.split_current_layer()
            self.visit_by_attr(node, "body")

    visit_AsyncFor = visit_For

    def visit_Module(self, node):
        self.generic_visit(node)

    def visit_ClassDef(self, node):
        # var assignments in class defs are not safe
        raise NotImplementedError

    def visit_FunctionDef(self, node):
        with self.new_layer_ctx(node.body):
            self.split_current_layer()
            self.visit_by_attr(node, "body")

    visit_AsyncFunctionDef = visit_FunctionDef

    def may_have_side_effects(self, node):
        return not self.no_side_effects_test(node)

    def visit_Assign(self, node):
        self.visit_by_attr(node, "value")
        for target in node.targets:
            if self.may_have_side_effects(target):
                self.split_current_layer(after_current_line=True)
                break

    def visit_AugAssign(self, node):
        self.visit_by_attr(node, "value")
        if self.may_have_side_effects(node.target):
            self.split_current_layer(after_current_line=True)

    def visit_Assert(self, node):
        self.visit_by_attr(node, "test")
        if self.may_have_side_effects(node.test):
            self.split_current_layer(after_current_line=True)

    def not_implemented(self, node):
        raise NotImplementedError

    visit_Nonlocal = visit_Global = not_implemented

    if PY_VERSION >= (3, 10):

        visit_Match = not_implemented
